use std::ptr;

pub trait EditableText {
    // pos is in utf8 codepoints
    fn insert_at(&mut self, pos: usize, contents: &str);
    fn remove_at(&mut self, pos: usize, length: usize);
}

impl EditableText for String {
    fn insert_at(&mut self, char_pos: usize, contents: &str) {
        // If you try to write past the end of the string for now I'll just write at the end.
        // Panicing might be a better policy.
        let byte_pos = self.char_indices().nth(char_pos)
            .map(|(p, _)| p).unwrap_or_else(|| self.len());
        //println!("pos {}", byte_pos);
        //self.insert_str(byte_pos, contents);

        let old_len = self.len();
        let new_bytes = contents.len();

        // This didn't work because it didn't change the string's length
        //self.reserve(new_bytes);

        // This is sort of ugly but its fine.
        for _ in 0..new_bytes { self.push('\0'); }

        //println!("new bytes {} {} {}", new_bytes, byte_pos, self.len() - byte_pos);
        unsafe {
            let bytes = self.as_mut_vec().as_mut_ptr();
            //println!("{:?}", self.as_mut_vec());
            ptr::copy(
                bytes.add(byte_pos),
                bytes.add(byte_pos + new_bytes),
                old_len - byte_pos
            );
            ptr::copy_nonoverlapping(
                contents.as_ptr(),
                bytes.add(byte_pos),
                new_bytes
            );
            //println!("{:?}", self.as_mut_vec());
        }
    }
    fn remove_at(&mut self, pos: usize, length: usize) {
        let byte_range = {
            let mut iter = self.char_indices().map(|(p, _)| p).skip(pos).peekable();

            let start = iter.peek().map_or_else(|| self.len(), |&p| p);
            let mut iter = iter.skip(length).peekable();
            let end = iter.peek().map_or_else(|| self.len(), |&p| p);

            start..end
        };

        self.drain(byte_range);
    }
}



#[cfg(test)]
mod tests {
    use super::EditableText;

    #[test]
    fn insert_simple() {
        let mut s = "".to_string();
        s.insert_at(0, "hi");
        assert_eq!(s, "hi");

        let mut s = "a".to_string();
        s.insert_at(0, "hi");
        assert_eq!(s, "hia");

        let mut s = "a".to_string();
        s.insert_at(1, "hi");
        assert_eq!(s, "ahi");

        let mut s = "ac".to_string();
        s.insert_at(1, "b");
        assert_eq!(s, "abc");
    }

    #[test]
    fn insert_unicode() {
        // I mean, its all unicode but ....
        let mut s = "𝄞𝄞".to_string();
        s.insert_at(0, "à");
        assert_eq!(s, "à𝄞𝄞");
        s.insert_at(2, "ë");
        assert_eq!(s, "à𝄞ë𝄞");
        s.insert_at(4, "ç");
        assert_eq!(s, "à𝄞ë𝄞ç");
        s.insert_at(6, "𝒲");
        assert_eq!(s, "à𝄞ë𝄞ç𝒲");
    }

    #[test]
    fn remove_simple() {
        let mut s = "à".to_string();
        s.remove_at(0, 1);
        assert_eq!(s, "");
        s.remove_at(0, 0);
        assert_eq!(s, "");

        let mut s = "à𝄞ç".to_string();
        s.remove_at(0, 1);
        assert_eq!(s, "𝄞ç");
        s.remove_at(1, 1);
        assert_eq!(s, "𝄞");
        s.remove_at(0, 1);
        assert_eq!(s, "");
    }
}
